from typing import override as _override

from leads import ShadowDevice as _ShadowDevice


class Pedal(_ShadowDevice):
    @_override
    def loop(self) -> None:
        pass

    @_override
    def read(self) -> None:
        pass



from typing import Optional as _Optional

from leads import L, device, controller, MAIN_CONTROLLER, get_controller, WHEEL_SPEED_CONTROLLER, SRWDataContainer, \
    DRWDataContainer, DC_MOTOR_CONTROLLER_A, LEFT_FRONT_WHEEL_SPEED_SENSOR, RIGHT_FRONT_WHEEL_SPEED_SENSOR, \
    CENTER_REAR_WHEEL_SPEED_SENSOR, LEFT_REAR_WHEEL_SPEED_SENSOR, RIGHT_REAR_WHEEL_SPEED_SENSOR, get_config
from leads.comm import Callback, Service, ConnectionBase
from leads_arduino import ArduinoMicro, WheelSpeedSensor
from leads_gui import Config
from leads_raspberry_pi import RaspberryPi4B, DCMotorController, VoltageSensor

config = get_config(Config)
BAUD_RATE: int = config.get("baud_rate", 9600)
WHEEL_SPEED_CONTROLLER_PORT: str = config.get("wheel_speed_controller_port", "COM3")
FRONT_WHEEL_DIAMETER: float = config.get("front_wheel_diameter", 20)  # 20 inches
REAR_WHEEL_DIAMETER: float = config.get("rear_wheel_diameter", 20)  # 20 inches
THROTTLE_PEDAL_PIN: int = config.get("throttle_pedal_pin", 2)
BRAKE_PEDAL_PIN: int = config.get("brake_pedal_pin", 3)
VOLTAGE_SENSOR_PIN: int = config.get("voltage_sensor_pin", 4)


@controller(MAIN_CONTROLLER)
class VeCController(RaspberryPi4B):
    def read(self) -> SRWDataContainer | DRWDataContainer:
        r = get_controller(WHEEL_SPEED_CONTROLLER).read()
        return SRWDataContainer(*r) if config.srw_mode else DRWDataContainer(*r)


class WheelSpeedControllerCallback(Callback):
    def on_connect(self, service: Service, connection: ConnectionBase) -> None:
        L.debug("Wheel speed controller connected")

    def on_receive(self, service: Service, msg: bytes) -> None:
        get_controller(WHEEL_SPEED_CONTROLLER).update(msg.decode())

    def on_fail(self, service: Service, error: Exception) -> None:
        L.error("Wheel speed controller error: " + str(error))


@controller(WHEEL_SPEED_CONTROLLER, MAIN_CONTROLLER, (
        WHEEL_SPEED_CONTROLLER_PORT, WheelSpeedControllerCallback(), BAUD_RATE
))
class WheelSpeedController(ArduinoMicro):
    def update(self, data: str) -> None:
        for d in self.devices():
            if data.startswith(d.tag()):
                d.update(float(data[data.find(":") + 1:]))

    def read(self) -> [float, float, float, _Optional[float]]:
        lfws = self.device(LEFT_FRONT_WHEEL_SPEED_SENSOR).read()
        rfws = self.device(RIGHT_FRONT_WHEEL_SPEED_SENSOR).read()
        rws = (self.device(CENTER_REAR_WHEEL_SPEED_SENSOR).read(),) if config.srw_mode else (
            self.device(LEFT_REAR_WHEEL_SPEED_SENSOR).read(),
            self.device(RIGHT_REAR_WHEEL_SPEED_SENSOR).read()
        )
        return min(lfws, rfws, *rws), (lfws + rfws) * .5, *rws


@device(*((
        (
                LEFT_FRONT_WHEEL_SPEED_SENSOR,
                RIGHT_FRONT_WHEEL_SPEED_SENSOR,
                CENTER_REAR_WHEEL_SPEED_SENSOR
        ), WHEEL_SPEED_CONTROLLER, [
            (FRONT_WHEEL_DIAMETER,),
            (FRONT_WHEEL_DIAMETER,),
            (REAR_WHEEL_DIAMETER,)
        ]
) if config.srw_mode else (
        (
                LEFT_FRONT_WHEEL_SPEED_SENSOR,
                RIGHT_FRONT_WHEEL_SPEED_SENSOR,
                LEFT_REAR_WHEEL_SPEED_SENSOR,
                RIGHT_REAR_WHEEL_SPEED_SENSOR
        ), WHEEL_SPEED_CONTROLLER, [
            (FRONT_WHEEL_DIAMETER,),
            (FRONT_WHEEL_DIAMETER,),
            (REAR_WHEEL_DIAMETER,),
            (REAR_WHEEL_DIAMETER,)
        ]
)))
class WheelSpeedSensors(WheelSpeedSensor):
    pass


# @device((THROTTLE_PEDAL, BRAKE_PEDAL), MAIN_CONTROLLER, [(THROTTLE_PEDAL_PIN, False), (BRAKE_PEDAL_PIN, True)])
# class Pedals(Pedal):
#     def __init__(self, pin: int, brake: bool) -> None:
#         super().__init__(pin)
#         self._brake: bool = brake


@device(DC_MOTOR_CONTROLLER_A, MAIN_CONTROLLER)
class DriverMotorController(DCMotorController):
    pass


@device("mvs", MAIN_CONTROLLER, (4,))
class MotorVoltageSensor(VoltageSensor):
    pass


_ = None  # null export



from abc import ABCMeta as _ABCMeta
from json import dumps as _dumps
from time import time as _time
from typing import override as _override


class Serializable(object):
    def to_dict(self) -> dict:
        attributes = dir(self)
        r = {}
        for n in attributes:
            if n.startswith("_"):
                continue
            v = self.__getattribute__(n)
            if type(v) in (int, float, str):
                r[n] = v
        return r


class DataContainer(Serializable, metaclass=_ABCMeta):
    def __init__(self, min_speed: int | float) -> None:
        self._time_stamp: int = int(_time() * 1000)
        self.speed: int | float = min_speed

    @_override
    def __str__(self) -> str:
        return _dumps(self.to_dict())

    def reset_time_stamp(self) -> None:
        """
        Reset the data's time stamp to now.
        """
        self._time_stamp = int(_time() * 1000)

    def get_time_stamp(self) -> int:
        """
        Get the data's time stamp.
        :return: the time stamp (ms)
        """
        return self._time_stamp

    @_override
    def to_dict(self) -> dict:
        """
        Convert the data into a dictionary.
        :return: a dictionary that contains all custom attributes of the container
        """
        (r := super().to_dict())["t"] = self._time_stamp
        return r

    def encode(self) -> bytes:
        """
        Encode the data into bytes for network transaction purposes.
        :return: JSON in bytes
        """
        return str(self).encode()


class SRWDataContainer(DataContainer):
    def __init__(self,
                 min_speed: int | float = 0,
                 front_wheel_speed: int | float = 0,
                 rear_wheel_speed: int | float = 0,
                 ) -> None:
        super().__init__(min_speed)
        self.front_wheel_speed: int | float = front_wheel_speed
        self.rear_wheel_speed: int | float = rear_wheel_speed


class DRWDataContainer(DataContainer):
    def __init__(self,
                 min_speed: int | float = 0,
                 front_wheel_speed: int | float = 0,
                 left_rear_wheel_speed: int | float = 0,
                 right_rear_wheel_speed: int | float = 0,
                 ) -> None:
        super().__init__(min_speed)
        self.front_wheel_speed: int | float = front_wheel_speed
        self.left_rear_wheel_speed: int | float = left_rear_wheel_speed
        self.right_rear_wheel_speed: int | float = right_rear_wheel_speed



from math import pi as _pi
from time import time as _time
from typing import override as _override

from leads import Device as _Device


def rpm2kmh(rpm: float, wheel_circumference: float) -> float:
    return rpm * wheel_circumference * .0006


class WheelSpeedSensor(_Device):
    def __init__(self, wheel_diameter: int | float) -> None:
        super().__init__()
        self._wheel_circumference: float = wheel_diameter * 2.54 * _pi
        self._wheel_speed: int | float = 0
        self._last_valid: float = 0

    @_override
    def update(self, data: int | float) -> None:
        self._wheel_speed = rpm2kmh(data, self._wheel_circumference)
        self._last_valid = _time()

    @_override
    def read(self) -> float:
        # add .001 to avoid zero division
        d = rpm2kmh(60 / (.001 + _time() - self._last_valid), self._wheel_circumference)
        return 0 if d < 1 else d if d < 5 else self._wheel_speed



from typing import Any as _Any

from leads.context import Context

type RequiredData = list[str] | tuple[list[str], list[str]]


class Plugin(object):
    def __init__(self, required_data: RequiredData) -> None:
        """
        :param required_data: required data or (required data in srw mode, required data in drw mode)
        """
        self._required_data: RequiredData = required_data
        self.state: dict[str, _Any] = {}
        self.enabled: bool = True
        self._context: Context | None = None

    def __getitem__(self, key: str) -> _Any:
        return self.state[key]

    def __setitem__(self, key: str, value: _Any) -> None:
        self.state[key] = value

    def bind_context(self, context: Context) -> None:
        self._context = context

    def _require_context(self) -> Context:
        if self._context:
            return self._context
        raise RuntimeError("No context bound")

    def required_data(self) -> list[str]:
        return (self._required_data if isinstance(self._required_data, list)
                else self._required_data[0] if self._require_context().srw_mode()
        else self._required_data[1])

    def load(self) -> None:
        self.on_load(self._require_context())

    def on_load(self, context: Context) -> None:
        pass

    def update(self, kwargs: dict[str, _Any]) -> None:
        self.on_update(self._require_context(), kwargs)

    def on_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        pass
